from enum import IntEnum
import json
import os
from typing import List
import yara
from ludvig.rules import RuleSetSource
import hashlib


class ConfigEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, RuleSetSource):
            return obj.__dict__
        return json.JSONEncoder.default(self, obj)


class Config:
    def __init__(
        self, config_path: str, rule_sources: List[RuleSetSource] = None
    ) -> None:
        self.config_path = config_path
        self.compiled_rules = os.path.join(config_path, "ludvig.rules")
        self.config_file = os.path.join(config_path, "config.json")
        if rule_sources and "Built-in" not in rule_sources:
            rule_sources.append(
                RuleSetSource(
                    "Built-In",
                    "secrets",
                    "https://github.com/FrodeHus/ludvig-rules/archive/refs/tags/v0.0.1.tar.gz",
                )
            )
        else:
            rule_sources = [
                RuleSetSource(
                    "Built-In",
                    "secrets",
                    "https://github.com/FrodeHus/ludvig-rules/archive/refs/tags/v0.0.1.tar.gz",
                )
            ]
        self.rule_sources = rule_sources

    def save(self):
        with open(self.config_file, "w") as f:
            f.write(json.dumps(self.__dict__, indent=4, cls=ConfigEncoder))

    @staticmethod
    def load():
        config_path = os.path.join(os.path.expanduser("~"), ".ludvig")
        if not os.path.exists(config_path):
            os.makedirs(config_path)
        config_file = os.path.join(config_path, "config.json")
        if os.path.exists(config_file):
            with open(config_file, "r") as f:
                config = json.loads(f.read())
                return Config(
                    config["config_path"],
                    Config.parse_rule_sets(config["rule_sources"]),
                )

        return Config(config_path)

    @staticmethod
    def parse_rule_sets(d: dict):
        rules = []
        for rule_source in d:
            rules.append(
                RuleSetSource(
                    rule_source["name"], rule_source["category"], rule_source["url"]
                )
            )


class Severity(IntEnum):
    UNKNOWN = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    CRITICAL = 4


class RuleMatch:
    def __init__(
        self,
        id: str,
        rule_name: str,
        severity: Severity = Severity.MEDIUM,
        category: str = None,
        description: str = None,
        tags: List[str] = None,
    ) -> None:
        self.id = id
        self.rule_name = rule_name
        self.severity = severity
        self.tags = tags
        self.category = category
        self.description = description


class YaraRuleMatch(RuleMatch):
    def __init__(self, yara: yara.Match) -> None:
        super().__init__(
            yara.meta["id"] if "id" in yara.meta else "LS00000",
            yara.rule,
            yara.meta["severity"] if "severity" in yara.meta else "UNKNOWN",
            yara.namespace,
            yara.meta["description"] if "description" in yara.meta else None,
            yara.tags,
        )


class FindingSample:
    def __init__(
        self, content: str, offset: int, deobfuscated=False, line_number: int = -1
    ) -> None:
        self.offset = offset
        self.line_number = line_number
        content = content[:10] + "..." if len(content) > 10 else content
        if deobfuscated:
            self.content = content
        else:
            obfuscated = "*" * len(content)
            self.content = (
                obfuscated[:10] + "..." if len(obfuscated) > 10 else obfuscated
            )

    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)

    @classmethod
    def from_yara_match(
        cls, match: yara.Match, deobfuscated=False, line_number: int = -1
    ) -> List["FindingSample"]:
        samples = []
        for str_match in match.strings:
            offset = str_match[0]
            identifier = str_match[1]
            data = str_match[2]
            if data.isascii():
                data = data.decode("utf-8")
            else:
                data = "".join(format(x, "02x") for x in data)
            samples.append(FindingSample(data, offset, deobfuscated, line_number))
        return samples


class Finding:
    def __init__(
        self,
        category: str,
        match: RuleMatch,
        samples: List[FindingSample],
        filename: str,
    ) -> None:
        self.name = "{}/{}".format(category, match.rule_name)
        self.match = match
        self.filename = filename
        self.samples = samples
        self.severity = match.severity
        self.comment = None
        self.properties = {category: category}
        self.hash = hashlib.sha1(
            "|".join(
                [self.name, self.filename, "@".join([s.content for s in self.samples])]
            ).encode()
        ).hexdigest()


class SecretFinding(Finding):
    def __init__(
        self,
        rule: RuleMatch,
        samples: List[FindingSample],
        filename: str,
        **kwargs,
    ) -> None:
        super().__init__(rule.category, rule, samples, filename)
        for arg in kwargs:
            self.properties[arg] = kwargs[arg]


class FindingEncoder(json.JSONEncoder):
    def default(self, o: Finding):
        return o.__dict__



from typing import List
import yara, os, glob
from ludvig.types import Finding, Severity
from ludvig.providers import FileSystemProvider
from ._common import BaseScanner


class FilesystemScanner(BaseScanner):
    def __init__(
        self,
        provider: FileSystemProvider,
        severity_level: Severity = Severity.MEDIUM,
        deobfuscated=False,
    ) -> None:
        super().__init__(deobfuscated)
        self.provider = provider
        self.severity_level = severity_level
        self.findings: List[Finding] = []

    def scan(self):
        for file_data, filename, properties in self.provider.get_files():
            findings = self.scan_file_data(file_data, filename, **properties)
            self.register_findings(findings)



from ludvig.scanners import ImageScanner
from ludvig.types import Severity
from ludvig.providers import ContainerProvider


def scan(
    repository: str,
    severity_level: Severity = Severity.MEDIUM,
    deobfuscated=False,
    output_sarif=None,
    include_first_layer=False,
    max_file_size=10000,
):
    """
    Scans a container image
    :param repository: Container image to scan (ex: myimage:1.1)
    :param severity_level: Set severity level for reporting
    :param deobfuscated: Returns any secrets found in plaintext. Default: False.
    :param output_sarif: Generates SARIF report if filename is specified.
    :param include_first_layer: Scan first layer (base image) as well - may affect speed. Default: False.
    :param max_file_size: Max file size for scanning (in bytes).
    """
    if isinstance(severity_level, str):
        severity_level = Severity[severity_level]
    provider = ContainerProvider(
        repository, include_first_layer, max_file_size=max_file_size
    )
    scanner = ImageScanner(provider, severity_level, deobfuscated)
    scanner.scan()
    if output_sarif:
        from ludvig.outputs import SarifConverter

        report = SarifConverter.from_findings(scanner.get_unique_findings())
        with open(output_sarif, "w") as r:
            r.write(report)
    return scanner.get_unique_findings()


def list_whiteouts(repository: str, include_first_layer=False):
    """
    Scans a container image for deleted files
    :param repository: Container image to scan
    :param include_first_layer: Scan first layer (base image) as well - may affect speed. Defaults to False.
    """
    provider = ContainerProvider(repository, include_first_layer)
    scanner = ImageScanner(provider)
    return scanner.list_whiteout()


def extract_file(repository: str, filename: str, output_file: str):
    """
    Extracts a file from the specified container image (even if marked as deleted) - first occurence only
    :param repository: Container image to read from
    :param filename: File to extract (full path)
    :param output_file: Output file
    """
    provider = ContainerProvider(repository)
    scanner = ImageScanner(provider)
    scanner.extract_file(filename, output_file)



from typing import Tuple


def get_line_number(content: str, position: int) -> Tuple[int, str]:
    lines = content.splitlines()
    lower = 0
    upper = len(lines) - 1
    while lower <= upper:
        index = int(lower + (upper - lower) / 2)
        start = content.index(lines[index])
        if position == start:
            return index, lines[index - 1]
        if start > position:
            upper = index - 1
        else:
            lower = index + 1

    return lower - 1, lines[lower - 1]



import json
from typing import IO
from cyclonedx.model import HashType
from cyclonedx.model.component import Component
from cyclonedx.parser import BaseParser, ParserWarning
from packageurl import PackageURL

from ludvig.sbom.dotnet.dependencies_file import DependenciesFile


class DotnetDependenciesParser(BaseParser):
    def __init__(self, dependency_content: str) -> None:
        super().__init__()

        deps = DependenciesFile.from_string(dependency_content)
        for lib in deps.libraries:
            # hashes = map(HashType.from_composite_str(lib.sha512))
            self._components.append(
                Component(
                    name=lib.name,
                    version=lib.version,
                    purl=PackageURL(type="nuget", name=lib.name, version=lib.version),
                )
            )



import tarfile
from typing import IO, List
from ._docker._definitions import Image, LayerFile
from ._providers import BaseFileProvider
from ._docker._main import read_local_docker_image
from knack.log import get_logger

logger = get_logger(__name__)


class ContainerProvider(BaseFileProvider):
    def __init__(
        self,
        repository: str,
        include_first_layer=False,
        exclusions: List[str] = None,
        max_file_size=10000,
    ) -> None:
        super().__init__(exclusions=exclusions, max_file_size=max_file_size)
        self.repository = repository
        self.include_first_layer = include_first_layer

    def get_files(self):
        with self.__get_image() as image:
            layers = image.layers[1:] if not self.include_first_layer else image.layers
            for layer in [l for l in layers if not l.empty_layer]:
                logger.info("layer %s: %s", layer.id, layer.created_by)
                with image.image_archive.extractfile(
                    "{}/layer.tar".format(layer.id)
                ) as layer_archive:
                    with tarfile.open(fileobj=layer_archive, mode="r") as lf:
                        try:
                            for member in lf.getmembers():
                                if (
                                    self.is_excluded(member.name)
                                    or not member.isfile
                                    or member.size > self.max_file_size
                                ):
                                    continue

                                file_data = self.__extract_file(lf, member)
                                if not file_data:
                                    continue
                                with LayerFile(file_data) as file_data:
                                    if not file_data:
                                        continue
                                    yield file_data, member.name, layer.id, layer.created_by
                        except tarfile.ReadError:
                            logger.error("failed to read files from layer %s", layer.id)

    def __get_image(self) -> Image:
        return read_local_docker_image(self.repository)

    def __extract_file(
        self, image: tarfile.TarFile, file: tarfile.TarInfo
    ) -> IO[bytes]:
        if file.isfile():
            return image.extractfile(file)
        return None



from typing import IO, List
import os
from ludvig.rules import load_yara_rules
from ludvig.types import Finding, FindingSample, SecretFinding, Severity, YaraRuleMatch
from ludvig.rules import download_rules
from ludvig import current_config
from knack.log import get_logger

logger = get_logger(__name__)


class BaseScanner:
    def __init__(self, deobfuscated=False) -> None:
        self.deobfuscated = deobfuscated
        if not os.path.exists(current_config.compiled_rules):
            logger.info("no rules found - downloading defaults...")
            download_rules(current_config.rule_sources, current_config.config_path)

        yara_rules = load_yara_rules(current_config.compiled_rules)
        self.__yara_rules = yara_rules
        self.__findings: List[Finding] = []

    def scan_file_data(
        self, file_data: IO[bytes], file_name: str, **kwargs
    ) -> List[Finding]:
        findings = []
        try:
            matches = self.__yara_rules.match(data=file_data.read())
            for match in matches:

                severity = (
                    Severity[match.meta["severity"]]
                    if "severity" in match.meta
                    else Severity.UNKNOWN
                )
                if severity < self.severity_level:
                    continue
                line = self.find_match_line_num(file_data, match)
                samples = FindingSample.from_yara_match(match, self.deobfuscated, line)
                finding = SecretFinding(
                    YaraRuleMatch(match), samples, file_name, **kwargs
                )
                findings.append(finding)
        except Exception as ex:
            return print(ex)
        finally:
            return findings

    def find_match_line_num(self, file_data, match):
        file_data.seek(0)
        fd = file_data.read(match.strings[0][0])
        line = fd.count(b"\n") + 1
        return line

    def register_findings(self, findings: List[Finding]):
        unique_hashes = {f.hash for f in self.findings}
        for finding in findings:
            if finding.hash in unique_hashes:
                continue
            self.__findings.append(finding)

    def get_unique_findings(self):
        unique_hashes = list({f.hash for f in self.__findings})
        unique_findings = []
        for finding in self.__findings:
            if finding.hash in unique_hashes:
                unique_findings.append(finding)
                unique_hashes.remove(finding.hash)

        return unique_findings



from ludvig.rules import download_rules, RuleSetSource
from ludvig import current_config


def download():
    """
    Downloads and compiles the configured rule sets.
    """
    download_rules(current_config.rule_sources, current_config.config_path)


def add_repo(name: str, category: str, url: str):
    """
    Adds a custom repository of YARA rules (needs to be in tar-gzipped format)
    """
    from ludvig import current_config

    current_config.rule_sources.append(RuleSetSource(name, category, url))
    current_config.save()



import glob
from tempfile import TemporaryDirectory, TemporaryFile
from typing import IO, List
import yara
import os
from ._definitions import RuleSetSource
import tarfile
import requests
from knack.log import get_logger

logger = get_logger(__name__)


def load_yara_rules(file: str) -> yara.Rules:
    return yara.load(file)


def download_rules(sources: List[RuleSetSource], config_path: str) -> yara.Rules:
    namespaces = {}
    for src in sources:
        set = __download_rule_set(src)
        if not set:
            continue
        ns = src.category.lower()
        if ns in namespaces:
            namespaces[ns] = namespaces[ns] + "\r\n" + set
        else:
            namespaces[ns] = set

    rules = yara.compile(sources=namespaces)
    output_file = os.path.join(config_path, "ludvig.rules")
    rules.save(output_file)


def __download_rule_set(source: RuleSetSource):
    name = source.name
    url = source.url
    logger.debug("downloading %s from %s", name, url)
    try:
        res = requests.get(url)

        with TemporaryFile(suffix=".tar.gz") as f:
            f.write(res.content)
            f.seek(0)
            return __extract_rules(f)
    except Exception as ex:
        logger.error("failed to retrieve rule package %s from %s: %s", name, url, ex)


def __extract_rules(rule_package: IO[bytes]):
    with TemporaryDirectory() as tmp_dir:
        archive = tarfile.open(fileobj=rule_package, mode="r:gz")
        archive.extractall(tmp_dir)
        rule_codes = []

        for file in glob.iglob(os.path.join(tmp_dir, "**/*.yar"), recursive=True):
            full_path = os.path.join(tmp_dir, file)
            with open(full_path, "r") as f:
                rule_code = f.read()
                rule_codes.append(rule_code)

        return "\r\n".join(rule_codes)



import base64
import os
import tarfile, re
from typing import IO, List
from ludvig.types import Finding, Severity
from ludvig.providers import ContainerProvider
from ._common import BaseScanner
from knack.log import get_logger

logger = get_logger(__name__)


class ImageScanner(BaseScanner):
    def __init__(
        self,
        file_provider: ContainerProvider,
        severity_level: Severity = Severity.MEDIUM,
        deobfuscated=False,
    ) -> None:
        super().__init__(deobfuscated)
        self.file_provider = file_provider
        self.severity_level = severity_level
        self.findings: List[Finding] = []

    def list_whiteout(self):
        whiteouts = []
        for (
            _,
            file_name,
            layer_id,
            layer_created_by,
        ) in self.file_provider.get_files():
            if os.path.basename(file_name).startswith(".wh."):
                whiteouts.append(
                    {
                        "layer": layer_id,
                        "created_by": layer_created_by,
                        "filename": file_name.replace(".wh.", ""),
                    }
                )

        return whiteouts

    def extract_file(self, filename: str, output: str):
        for (
            file,
            file_name,
            _,
            _,
        ) in self.file_provider.get_files():
            if file_name.lower() == filename:
                logger.info("found %s - extracting to %s ...", filename, output)
                with open(output, "wb") as f:
                    f.write(file.file_data.read())
                return

    def scan(self):
        for (
            file,
            file_name,
            _,
            layer_created_by,
        ) in self.file_provider.get_files():
            if os.path.basename(file_name).startswith(".wh."):
                self.__whiteout(file_name, layer_created_by)
            findings = self.scan_file_data(
                file.file_data, file_name, docker_instruction=layer_created_by
            )
            self.register_findings(findings)

    def __whiteout(self, filename: str, layer_created_by: str):
        finding = [
            finding
            for finding in self.findings
            if finding.filename == filename.replace(".wh.", "")
        ]

        for f in finding:
            f.properties["removed_by"] = layer_created_by
            f.properties["whiteout"] = True

    def __scan_environment(self, variables: List[str]) -> Finding:
        pass

    def __decode_content(self, content: str) -> str:
        for match in self.__possible_base64_encoding(content):
            try:
                decoded = base64.b64decode(match.group()).decode("utf-8")
                content = content.replace(match.group(), decoded)
            except UnicodeDecodeError:
                continue
        return content

    def __possible_base64_encoding(self, content: str):
        return re.finditer(
            r"(^|\s+)[\"']?((?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{4}|[A-Za-z0-9+\/]{3}=|[A-Za-z0-9+\/]{2}={2}))[\"']",
            content,
            flags=re.RegexFlag.MULTILINE,
        )



import os
from unittest import TestCase
from ludvig.providers._git import GitPackIndex, GitPack, GitRepository


class TestParseGitHistory(TestCase):
    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)

    def test_parse_pack_index(self):
        idx_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.idx")
        idx = GitPackIndex(idx_file)
        self.assertEqual(len(idx["objects"]), idx["total_objects"])
        self.assertEqual(len(idx["objects"]), 11)

    def test_parse_pack(self):
        idx_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.idx")
        idx = GitPackIndex(idx_file)
        pack_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.pack")
        with GitPack(pack_file, idx) as pack:
            self.assertEqual(pack.entries["total_objects"], pack.idx["total_objects"])

    def test_parse_commit(self):
        idx_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.idx")
        idx = GitPackIndex(idx_file)
        pack_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.pack")
        with GitPack(pack_file, idx) as pack:
            commit = pack.get_commit_by_hash("04ac220976f58bfc67a64bcb3b30571eace538cf")
            self.assertEqual(
                commit.tree_hash, "ca38fde4b9ebe4001389c96176e253469e2bb0d5"
            )

    def test_retrieve_object_offset(self):
        idx_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.idx")
        idx = GitPackIndex(idx_file)
        offset = idx.get_offset("04ac220976f58bfc67a64bcb3b30571eace538cf")
        self.assertEqual(offset, 12)

    def test_parse_git_tree(self):
        pack_file = os.path.join(os.path.dirname(__file__), "fixtures/pack")
        with GitRepository([pack_file]) as p:
            tree = p.get_tree(hash="b6683c89de7b124d3d268affb4811f4f37cb4c88")
            self.assertEqual(len(tree.leafs), 2)
            self.assertTrue([o.path for o in tree.leafs] == ["README.md", "test"])

    def test_apply_delta(self):
        idx_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.idx")
        idx = GitPackIndex(idx_file)
        pack_file = os.path.join(os.path.dirname(__file__), "fixtures/pack.pack")
        with GitPack(pack_file, idx) as pack:
            content, _, _ = pack.get_pack_object(
                "0d75dd8e0d10925f49d9e3e4f06d70fe17247198"
            )
            self.assertEqual(
                """{\n\t"title": "yaay a title",\n\t"pat": "ghp_123abc1234567890abcdefjiklmnopqrstuv",\n\t"server_access_token": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJodHRwczovL2R1bW15LnN0dWZmLmNvbS8iLCJpYXQiOjE1MTYyMzkwMjIsIm5iZiI6MTY3MzAzMjExNywiZXhwIjoxNjczMDM2NDIwLCJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIn0.-cDdFD5abX_VeowZzLSoEL4rEbiVlJYOd01Zhg5Ntrk"\n}\n""",
                content.decode("utf-8"),
            )

    def test_build_history(self):
        pack_file = os.path.join(os.path.dirname(__file__), "fixtures/pack")
        with GitRepository([pack_file]) as p:
            history = p.build_history()



from typing import List
import os
import fnmatch


class BaseFileProvider:
    def __init__(self, exclusions: List[str] = None, max_file_size=10000) -> None:
        self.exclusions = exclusions or []
        self.__read_ignore_file()
        self.max_file_size = max_file_size

    def get_files(self):
        pass

    def is_excluded(self, filename: str) -> bool:
        for exclude in self.exclusions:
            return fnmatch.fnmatch(filename, exclude)

    def __read_ignore_file(self):
        if os.path.exists(".ludvignore"):
            with open(".ludvignore", "r") as fp:
                ignore_data = fp.readlines()
                self.exclusions.extend(ignore_data)



from collections import OrderedDict


def transform_finding_list(result):
    transformed = []
    for r in result:
        res = OrderedDict(
            [
                ("RuleId", r["match"]["id"]),
                ("Name", r["match"]["ruleName"]),
                ("Severity", r["severity"]),
                ("Filename", r["filename"]),
                ("LineNumber", r["samples"][0]["lineNumber"]),
            ]
        )
        transformed.append(res)
    return transformed


def transform_git_finding_list(result):
    transformed = []
    for r in result:

        res = OrderedDict(
            [
                ("RuleId", r["match"]["id"]),
                ("Name", r["match"]["ruleName"]),
                ("Severity", r["severity"]),
                ("Filename", r["filename"]),
                ("LineNumber", r["samples"][0]["lineNumber"]),
                ("File SHA", r["properties"]["file_hash"]),
                ("Commit SHA", r["properties"]["commit_hash"]),
            ]
        )
        transformed.append(res)
    return transformed



from unittest import TestCase
from ludvig.types import Finding, RuleMatch, Severity
from ludvig.outputs import SarifConverter


class TestSarifLogGenerator(TestCase):
    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)

    def test_result_from_finding(self) -> None:
        finding = Finding(
            "secret",
            RuleMatch("GitHub Token", Severity.CRITICAL, "secrets", "Test rule"),
            None,
            "testfile.json",
        )
        report = SarifConverter.from_findings([finding])
        self.assertEqual(len(report.runs[0].results), 1)



from ludvig.types import Severity
from ludvig.providers import GitRepositoryProvider
from ludvig.scanners import FilesystemScanner
from knack.log import get_logger

logger = get_logger(__name__)


def scan(
    path: str,
    commit: str = None,
    fast: bool = False,
    severity_level: Severity = Severity.MEDIUM,
    deobfuscated=False,
    output_sarif=None,
    max_file_size=10000,
):
    """
    Scans the history of a Git repository
    :param path: Path to Git repository
    :param commit: Commit SHA to scan
    :param fast: Scans Git history as fast as possible - might lead to inaccurate results
    :param severity_level: Set severity level for reporting
    :param deobfuscated: Returns any secrets found in plaintext. Default: False.
    :param output_sarif: Generates SARIF report if filename is specified.
    :param max_file_size: Max file size for scanning (in bytes).
    """
    if isinstance(severity_level, str):
        severity_level = Severity[severity_level]

    if fast:
        logger.warn("fast scan is ENABLED - might lead to inaccurate results")

    git_provider = GitRepositoryProvider(path, commit=commit, fast_scan=fast)
    scanner = FilesystemScanner(git_provider, severity_level, deobfuscated)
    scanner.scan()
    if output_sarif:
        from ludvig.outputs import SarifConverter

        report = SarifConverter.from_findings(scanner.get_unique_findings())
        with open(output_sarif, "w") as r:
            r.write(report)
    return scanner.get_unique_findings()
